"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CtorEnsure = exports.getActiveControls = exports.registry = exports.META_KEY_DISPLAYNAME = exports.META_KEY_VALIDATION = void 0;
var tslib_1 = require("tslib");
var _1 = require(".");
var ctor_ensure_exception_1 = require("./ctor-ensure.exception");
var optionality_enum_1 = (0, tslib_1.__importDefault)(require("./optionality.enum"));
var util_1 = require("./util");
// Key used for metadata regarding validation
exports.META_KEY_VALIDATION = 'CTOR_ENSURE:VALIDATION';
exports.META_KEY_DISPLAYNAME = 'CTOR_ENSURE:DISPLAYNAME';
// Unique metadata valiation key
var META_KEY_VALIDATION_UNIQUE = function (displayname) { return "CTOR_ENSURE:" + displayname + ":VALIDATION"; };
// This registry saves the class' displayname to it's corresponding class
exports.registry = {};
// Key used to store blocked field information in class prototypes
var META_KEY_BLOCKED_FIELDS = 'CTOR_ENSURE:BLOCKED_FIELDS';
/**
 * Check wheter or not a value is conforming to it's optionality state
 * @param value Value to check
 * @param optionality Optionality-level to check against
 * @returns The error description if any, null if it passed and a
 * boolean, whether or not further validation is necessary
 */
var checkOptionality = function (value, optionality) {
    // Account for optionality
    switch (optionality) {
        case optionality_enum_1.default.NULLABLE:
            if (value === null)
                return [null, false];
            if (value !== undefined)
                return [null, true];
            break;
        case optionality_enum_1.default.OMITTABLE:
            if (value === undefined)
                return [null, false];
            if (value !== null)
                return [null, true];
            break;
        case optionality_enum_1.default.IRRELEVANT:
            return [null, !(value === null || value === undefined)];
        default:
            return [null, true];
    }
    return [(0, _1.template)('OPTIONALITY', {
            nullable: optionality === optionality_enum_1.default.NULLABLE,
        }), false];
};
/**
 * Validate an array of constructor arguments based on an array of controls which
 * define the requested validation schema, then collect and return those errors
 * @param args Constructor arguments to validate
 * @param controls Controls to use for the schema
 * @param multipleErrorsPerField Whether or not to exit after one error per field
 * @param argMap Argument map of field displayname to constructor value
 * @returns List of occurred errors
 */
var validateCtorArgs = function (args, controls, multipleErrorsPerField, argMap) {
    var errors = [];
    // Iterate every control
    for (var k = 0; k < controls.length; k += 1) {
        var currControl = controls[k];
        // Skip this control if callback returns with true
        if (currControl.skipOn) {
            var skip = currControl.skipOn(argMap);
            if (skip)
                continue;
        }
        // Get target value from constructor args
        var currArg = args[currControl.ctorInd];
        // Check optionality
        var _a = (0, tslib_1.__read)(checkOptionality(currArg, currControl.optional), 2), optErr = _a[0], needsValidation = _a[1];
        // Didn't pass optionality, raise error and continue
        // to next validated arg
        if (optErr) {
            errors.push({
                field: currControl.displayName,
                description: optErr,
                value: currArg,
            });
            continue;
        }
        // No further validation needed
        if (!needsValidation)
            continue;
        // Validate all values individually (to support arrays)
        var values = Array.isArray(currArg) ? currArg : [currArg];
        for (var j = 0; j < values.length; j += 1) {
            var currValue = values[j];
            // Flag, marks if the current value has passed validation
            var currValuePassed = true;
            // Validate whole config chain from top to bottom
            for (var i = 0; i < currControl.configs.length; i += 1) {
                var currConfig = currControl.configs[i];
                // Call ensure process callback with all dependencies
                var res = currConfig.process(currValue, controls, args, currControl, currArg);
                // Validation error occurred, push and set flag
                if (res.error) {
                    var err = {
                        field: currControl.displayName,
                        description: (0, _1.evalStrThunk)(currConfig.description),
                        value: res.value,
                    };
                    errors.push(err);
                    // Only one error per field is desired
                    if (!multipleErrorsPerField)
                        break;
                    currValuePassed = false;
                }
            }
            // One of the array's values didn't pass, skip all other values
            if (!currValuePassed && Array.isArray(currArg))
                break;
        }
    }
    return errors;
};
/**
 * Get all active controls from a class' metadata, taking block-lists into account
 * @param clazz Class to search in
 * @param displayname Displayname of the class to construct metadata key
 * @param blockedFields List of currently blocked fields
 * @returns List of validation-controls that are currently active
 */
var getActiveControls = function (clazz, displayname, blockedFields) {
    // Get controls from this class
    var controls = (Reflect.getOwnMetadata(META_KEY_VALIDATION_UNIQUE(displayname), clazz) || []);
    // All fields have been blocked (full disable)
    if (blockedFields.includes('*'))
        return [];
    // Return filtered list of controls
    return controls.filter(function (ctl) { return !blockedFields.some(function (blk) { return blk.toLowerCase() === ctl.displayName.toLowerCase(); }); });
};
exports.getActiveControls = getActiveControls;
/**
 * Decorator signalling that the following class' constructor will be validated
 * using {@link ValidatedArg} decorators on members in constructor.
 * All validation errors will be thrown using a single {@link CtorEnsureException}
 * @param displayname Displayname of class
 * @param multipleErrorsPerField Whether or not to display multiple errors per field, default false
 */
var CtorEnsure = function (config) { return function (Clazz) {
    // Buffer original clazz and it's prototype before interception
    var OrigClazz = Clazz;
    var origProto = OrigClazz.prototype;
    // Intercept constructor call
    // eslint-disable-next-line func-names
    var interceptor = function () {
        var ctorArgs = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            ctorArgs[_i] = arguments[_i];
        }
        // Create argument map, mapping displayname to actual value
        var argMap = (0, exports.getActiveControls)(interceptor, config.displayname, [])
            .reduce(function (acc, curr) {
            acc[curr.displayName] = ctorArgs[curr.ctorInd];
            return acc;
        }, {});
        // Evaluate skip callback if set
        var skip = config.skipOn ? config.skipOn(argMap) : false;
        // Most base-class will define blocked fields on most super-class' prototype
        var sC = (0, util_1.getLastSuperclassProto)(origProto);
        if (!Reflect.hasMetadata(META_KEY_BLOCKED_FIELDS, sC)) {
            Reflect.defineMetadata(META_KEY_BLOCKED_FIELDS, 
            // If inherited errors are also skipped, put * as blocklist
            (skip && config.skipOnSkipsInherited) ? ['*'] : config.blockInheritanceForFields || [], sC);
        }
        // Retrieve blocked fields
        var blockedFields = Reflect.getMetadata(META_KEY_BLOCKED_FIELDS, sC);
        // Block all fields for classes higher up the chain if inheritance has been broken
        if (!config.inheritValidation && !blockedFields.includes('*'))
            Reflect.defineMetadata(META_KEY_BLOCKED_FIELDS, (0, tslib_1.__spreadArray)((0, tslib_1.__spreadArray)([], (0, tslib_1.__read)(blockedFields), false), ['*'], false), sC);
        // Arrived at most super-class, remove definition
        if (sC === origProto)
            Reflect.deleteMetadata(META_KEY_BLOCKED_FIELDS, sC);
        // Get all currently active controls
        var controls = (0, exports.getActiveControls)(interceptor, config.displayname, blockedFields);
        // Validate args based on defined controls
        // Skip validation if skip resulted in true
        var errors = skip ? [] : validateCtorArgs(ctorArgs, controls, config.multipleErrorsPerField || false, argMap);
        try {
            // Call ctor, this will possibly throw super-call ensure-exceptions
            // Pass invocation information as last argument, that will be caught and removed later on
            var inst = new (OrigClazz.bind.apply(OrigClazz, (0, tslib_1.__spreadArray)([void 0], (0, tslib_1.__read)(ctorArgs), false)))();
            // No super-errors occurred, throw own errors, if any
            if (errors.length > 0)
                throw new ctor_ensure_exception_1.CtorEnsureException(interceptor, errors);
            // Return instance
            return inst;
        }
        catch (e) {
            // Listen for exceptions of super-classes
            if (e instanceof ctor_ensure_exception_1.CtorEnsureException && e.clazz !== interceptor)
                // Throw exception with changed class and merged errors
                throw new ctor_ensure_exception_1.CtorEnsureException(interceptor, (0, tslib_1.__spreadArray)((0, tslib_1.__spreadArray)([], (0, tslib_1.__read)(e.errors), false), (0, tslib_1.__read)(errors), false));
            // Just rethrow others
            throw e;
        }
    };
    // Unique-ify validation based on displayname
    // Used keep apart validation schemes, since inner decorators
    // don't know the displayname at their time of execution yet
    Reflect.defineMetadata(META_KEY_VALIDATION_UNIQUE(config.displayname), Reflect.getMetadata(exports.META_KEY_VALIDATION, Clazz), interceptor);
    Reflect.deleteMetadata(exports.META_KEY_VALIDATION, Clazz);
    // Copy prototype and static members
    Object.getOwnPropertyNames(Clazz)
        .filter(function (it) { var _a; return (_a = Object.getOwnPropertyDescriptor(Clazz, it)) === null || _a === void 0 ? void 0 : _a.writable; })
        .forEach(function (it) {
        interceptor[it] = Clazz[it];
    });
    // Copy existing metadata
    Reflect.getMetadataKeys(Clazz).forEach(function (key) { return Reflect.defineMetadata(key, Reflect.getMetadata(key, Clazz), interceptor); });
    // (Re-)Define display-name
    Reflect.defineMetadata(exports.META_KEY_DISPLAYNAME, config.displayname, interceptor);
    // Register this new class in the registry
    exports.registry[config.displayname] = interceptor;
    // This will be the new constructor
    return interceptor;
}; };
exports.CtorEnsure = CtorEnsure;
//# sourceMappingURL=ctor-ensure.decorator.js.map