"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.template = exports.processFunction = exports.escapeFunctionResult = exports.stripEscapes = exports.replaceVariables = exports.findFunctionCalls = exports.getRegisteredTemplateFunctions = exports.registerTemplateFunction = exports.key = void 0;
var tslib_1 = require("tslib");
var util_1 = require("./util");
// Helper formatter to get the full env key from a template name
var key = function (name) { return "CTOR_ENSURE_" + name.toUpperCase() + "_DESC"; };
exports.key = key;
// Default functions that are always available
var PREDEFINED_FUNCTIONS = {
    opt: util_1.strOpt,
    plur: util_1.pluralize,
    trn: util_1.ternaryString,
};
/**
 * Register a new function to be available globally inside all templates
 * @param name Name of the function inside template
 * @param func Reference to function
 */
var registerTemplateFunction = function (name, func) {
    PREDEFINED_FUNCTIONS[name] = func;
};
exports.registerTemplateFunction = registerTemplateFunction;
/**
 * Fetch all currently registered template functions
 * @returns A copy of the current function map
 */
var getRegisteredTemplateFunctions = function () { return ((0, tslib_1.__assign)({}, PREDEFINED_FUNCTIONS)); };
exports.getRegisteredTemplateFunctions = getRegisteredTemplateFunctions;
/**
 * Find function invocations within the template, only search for known functions
 * @param template Template to query
 * @param funcs Known functions to check for
 * @returns List of indices where function calls begin
 */
var findFunctionCalls = function (template, funcs) {
    var calls = [];
    // Function names sorted by length ascending to have shortest name as first element
    var funcNames = Object.keys(funcs).sort(function (it) { return it.length; });
    var _loop_1 = function (i) {
        // Iterate all known functions
        for (var j = 0; j < funcNames.length; j += 1) {
            // Beginning of a function invocation is <name>:
            var funcBegin = funcNames[j] + ":";
            if (
            // Check if function name and arg separator are found in conjunction starting from i
            funcBegin.split('').every(function (v, ci) { return template[i + ci] === v; }) &&
                // AND followed by either string delimiter " or var begin {
                (template[i + funcBegin.length] === '"' || template[i + funcBegin.length] === '{')) {
                calls.push(i);
                break;
            }
        }
    };
    // Iterate from tail to head, optimizing by subtracting shortest name form tail
    // adding two to it's length, because of arg separator and value delimiter
    for (var i = template.length - (funcNames[0].length + 2); i >= 0; i -= 1) {
        _loop_1(i);
    }
    return calls;
};
exports.findFunctionCalls = findFunctionCalls;
/**
 * Substitute known variable placeholders or remove escape sequences
 * @param template Template string to substitute in
 * @param vars Known variables to apply
 * @returns Output string with changed placeholders
 */
var replaceVariables = function (template, vars) {
    var result = template.slice();
    Object.keys(vars).forEach(function (varName) {
        // Replace all known variables
        result = result.slice();
        result = result.replace(new RegExp("(?<!\\\\)\\{" + varName + "\\}", 'g'), vars[varName]);
    });
    return result;
};
exports.replaceVariables = replaceVariables;
/**
 * Strip known escape sequences away, if they're not escaped twice
 * @param template Template string to substitute in
 * @returns Output string with stripped escapes for colons or curly brackets
 */
var stripEscapes = function (template) {
    var result = template.slice();
    // Strip all colon (arg-separator) escapes that are not escaped twice
    result = result.replace(/(?<!\\)\\:/g, ':');
    // Strip all quote escapes that are not escaped twice
    result = result.replace(/(?<!\\)\\"/g, '"');
    // Strip all opening curly bracket escapes that are not escaped twice
    result = result.replace(/(?<!\\)\\{/g, '{');
    // Strip all closing curly bracket escapes that are not escaped twice
    result = result.replace(/(?<!\\)\\}/g, '}');
    // Strip down multi-escapes to a single escape
    result = result.replace(/(\\+)(\\[:"{}]{1})/g, '$2');
    return result;
};
exports.stripEscapes = stripEscapes;
/**
 * Escape a string's critical symbols so that it can be used
 * within another function
 * @param input Input string to be escaped
 * @returns String safe to be used within another function
 */
var escapeFunctionResult = function (input, funcs, vars) {
    var result = input.slice();
    // Escape colon (arg-separator) right after a function invocation
    Object.keys(funcs).forEach(function (func) {
        result = result.replace(new RegExp("(?<=" + func + "):", 'g'), '\\:');
    });
    // Escape unescaped quotes
    result = result.replace(new RegExp('(?<!\\\\)"', 'g'), '\\"');
    // Escape curly brackets from variables
    Object.keys(vars).forEach(function (varName) {
        result = result.replace(new RegExp("(?<!\\\\)\\{" + varName + "(?!\\\\)\\}", 'g'), "\\{" + varName + "\\}");
    });
    return result;
};
exports.escapeFunctionResult = escapeFunctionResult;
/**
 * Process a function on the template string and return the result
 * @param defInd Index within template where the function starts
 * @param template Current state of the template
 * @param vars Variables to be used as function parameters
 * @param funcs Functions that are available for this template
 * @returns New template, after evaluating this function
 */
var processFunction = function (defInd, template, vars, funcs) {
    var result = template.slice();
    // Beginning of arguments is after first :, relative to definition
    var argsBegin = result.indexOf(':', defInd) + 1;
    // Figure out function name, between start and :
    var name = result.substring(defInd, argsBegin - 1);
    var argBuf = ''; // Argument buffer collecting characters over time
    var args = []; // List of arguments, filled by argBuf
    var inStr = false; // Whether or not currently within a string
    var varBegin = null; // Current variable begin, null means not within one
    var endInd = defInd; // Index of function completion, start=end for now
    var wasArgVar = false; // Flag if last substitution was a argument variable
    // Iterate till furthest possible point: EOL
    for (var i = argsBegin; i < result.length; i += 1) {
        var curr = result[i];
        var isEscaped = i !== 0 && result[i - 1] === '\\'; // Has current char been escaped? (prev was \)
        var isStrDelim = curr === '"' && !isEscaped; // Is current char a non-escaped string delimiter?
        // Set end at i, it will be last reached char, if
        // not otherwise overriden
        endInd = i;
        // Flip inStr state, if not next iteration after arg var substitution
        if (isStrDelim && !wasArgVar) {
            inStr = !inStr;
            // Last char ended string, push
            if (i === result.length - 1)
                args.push(argBuf);
            continue;
        }
        // Add current to arg buffer
        argBuf += curr;
        // Begin of variable
        if (curr === '{' && !isEscaped)
            varBegin = i;
        // Not inside a string or a variable, colon delimiter expected
        // Skip this check on first character of args
        if (!inStr && !varBegin && curr !== ':' && i !== argsBegin) {
            // End index is -1, since current is not part of the function anymore
            endInd = i - 1;
            // Push arg buffer if it wasn't a variable (is already in buffer)
            // or if it was escaped or string ended
            if (!wasArgVar && (curr === '\\' || result[i - 1] === '"'))
                args.push(argBuf.substring(0, argBuf.length - 1));
            // Handy function visualizer
            // console.log(`${result}\n${' '.repeat(defInd)}|${'-'.repeat(i - defInd - 2)}|`);
            break;
        }
        if (
        // Ignore if within variable processing
        !varBegin &&
            // Argument separator reached that is not inside string or variable
            (curr === ':' && !inStr && !isEscaped)) {
            // Remove separator from buffer
            argBuf = argBuf.substring(0, argBuf.length - 1);
            // Push and reset, if argument wasn't a variable
            if (!wasArgVar) {
                args.push(argBuf);
                argBuf = '';
            }
        }
        // A variable that has been started ends now
        if (varBegin !== null && curr === '}') {
            // Variable name is from begin + 1 to current - 1
            var varName = (result.substring(varBegin + 1, i));
            // Find the variable by it's name
            var val = vars[varName];
            // Unknown variable call, leave as is
            // Escaped variables have a trailing backslash varname\
            // and thus automatically become the desired value
            if (val === undefined)
                val = "{" + varName + "}";
            // String representation of value
            var strVal = String(val);
            // Substitute variable in result
            result = result.substring(0, varBegin) + strVal + result.substring(i + 1, result.length);
            // Next iteration will pick up right after variable substitution
            i = varBegin + strVal.length - 1;
            endInd = i + 1;
            // Variable done
            varBegin = null;
            // Variable as function-argument has been substituted
            if (!inStr) {
                // Clear argument buffer (holds placeholder)
                argBuf = '';
                // Push actual value
                args.push(val);
                // Set flag
                wasArgVar = true;
            }
            // Variable as plain text has been substituted, just replace in argument buffer
            else {
                argBuf = argBuf.substring(0, argBuf.length - (varName.length + 2)) + strVal;
            }
            continue;
        }
        // Clear flags
        wasArgVar = false;
    }
    // Find the function by it's name
    var func = funcs[name];
    if (!func)
        return template;
    var res = func.apply(void 0, (0, tslib_1.__spreadArray)([], (0, tslib_1.__read)(args), false));
    // No function result
    if (res === undefined)
        throw new Error('Function returned undefined value!');
    // Return input, where function invocation is substituted for function result
    // Every function needs to escape it's own result, to avoid re-interpret
    return result.substring(0, defInd) + (0, exports.escapeFunctionResult)(func.apply(void 0, (0, tslib_1.__spreadArray)([], (0, tslib_1.__read)(args), false)), funcs, vars) + result.substring(endInd + 1, result.length);
};
exports.processFunction = processFunction;
/**
 * Get the rendered template from a environment variable
 * @param name Name of the template
 * @param vars Variables that need to be available
 * @param funcs Custom added functions
 * @returns Rendered template with all functions and variables substituted
 */
var template = function (name, vars, funcs) {
    if (vars === void 0) { vars = null; }
    if (funcs === void 0) { funcs = null; }
    var templateString = process.env[(0, exports.key)(name)];
    // Template unknown
    if (!templateString)
        throw new SyntaxError("The template " + (0, exports.key)(name) + " is not registered as an ENV-VAR!");
    // Add custom functions to predefined functions
    var knownFunctions = (0, tslib_1.__assign)((0, tslib_1.__assign)({}, funcs), PREDEFINED_FUNCTIONS);
    var knownVariables = (0, tslib_1.__assign)({}, vars);
    // Find all function calls and process them
    (0, exports.findFunctionCalls)(templateString, knownFunctions).forEach(function (defInd) {
        templateString = (0, exports.processFunction)(defInd, templateString, knownVariables, knownFunctions);
    });
    templateString = (0, exports.replaceVariables)(templateString, knownVariables);
    templateString = (0, exports.stripEscapes)(templateString);
    // Return final template after replacing all non-function variables too
    // Trim template string to allow for easier notation within template
    return templateString.trim();
};
exports.template = template;
//# sourceMappingURL=description-template.factory.js.map